@@ -26,7 +26,6 @@ pub struct InstData {
     op: InstOpcode,
     operands: Vec<Operand>,
     clobbers: Vec<PReg>,
-    is_safepoint: bool,
 }
 
 impl InstData {
@@ -35,7 +34,6 @@ impl InstData {
             op: InstOpcode::Branch,
             operands: vec![],
             clobbers: vec![],
-            is_safepoint: false,
         }
     }
     pub fn ret() -> InstData {
@@ -43,7 +41,6 @@ impl InstData {
             op: InstOpcode::Ret,
             operands: vec![],
             clobbers: vec![],
-            is_safepoint: false,
         }
     }
 }
@@ -103,14 +100,6 @@ impl Function for Func {
         &self.block_params_out[block.index()][succ][..]
     }
 
-    fn requires_refs_on_stack(&self, insn: Inst) -> bool {
-        self.insts[insn.index()].is_safepoint
-    }
-
-    fn reftype_vregs(&self) -> &[VReg] {
-        &self.reftype_vregs[..]
-    }
-
     fn debug_value_labels(&self) -> &[(VReg, Inst, Inst, u32)] {
         &self.debug_value_labels[..]
     }
@@ -515,19 +504,12 @@ impl Func {
                     )));
                 }
 
-                let is_safepoint = opts.reftypes
-                    && operands
-                        .iter()
-                        .all(|op| !builder.f.reftype_vregs.contains(&op.vreg()))
-                    && bool::arbitrary(u)?;
-
                 builder.add_inst(
                     Block::new(block),
                     InstData {
                         op: InstOpcode::Op,
                         operands,
                         clobbers,
-                        is_safepoint,
                     },
                 );
                 avail.push(vreg);
@@ -583,9 +565,6 @@ impl Func {
 impl core::fmt::Debug for Func {
     fn fmt(&self, f: &mut core::fmt::Formatter) -> core::fmt::Result {
         write!(f, "{{\n")?;
-        for vreg in self.reftype_vregs() {
-            write!(f, "  REF: {}\n", vreg)?;
-        }
         for (i, blockrange) in self.blocks.iter().enumerate() {
             let succs = self.block_succs[i]
                 .iter()
@@ -620,9 +599,6 @@ impl core::fmt::Debug for Func {
                 i, params_in, succs, preds
             )?;
             for inst in blockrange.iter() {
-                if self.requires_refs_on_stack(inst) {
-                    write!(f, "    -- SAFEPOINT --\n")?;
-                }
                 write!(
                     f,
                     "    inst{}: {:?} ops:{:?} clobber:{:?}\n",
