@@ -17,7 +17,7 @@ use crate::cfg::CFGInfo;
 use crate::index::ContainerComparator;
 use crate::indexset::IndexSet;
 use crate::{
-    define_index, Allocation, Block, Edit, Function, FxHashSet, Inst, MachineEnv, Operand, PReg,
+    define_index, Allocation, Block, Edit, Function, FxHashSet, MachineEnv, Operand, PReg,
     ProgPoint, RegClass, VReg,
 };
 use alloc::collections::BTreeMap;
@@ -25,8 +25,7 @@ use alloc::string::String;
 use alloc::vec::Vec;
 use core::cmp::Ordering;
 use core::fmt::Debug;
-use hashbrown::{HashMap, HashSet};
-use smallvec::SmallVec;
+use smallvec::{smallvec, SmallVec};
 
 /// A range from `from` (inclusive) to `to` (exclusive).
 #[derive(Clone, Copy, Debug, PartialEq, Eq)]
@@ -94,11 +93,11 @@ impl core::cmp::Ord for CodeRange {
     }
 }
 
-define_index!(LiveBundleIndex);
-define_index!(LiveRangeIndex);
-define_index!(SpillSetIndex);
+define_index!(LiveBundleIndex, LiveBundles, LiveBundle);
+define_index!(LiveRangeIndex, LiveRanges, LiveRange);
+define_index!(SpillSetIndex, SpillSets, SpillSet);
 define_index!(UseIndex);
-define_index!(VRegIndex);
+define_index!(VRegIndex, VRegs, VRegData);
 define_index!(PRegIndex);
 define_index!(SpillSlotIndex);
 
@@ -123,8 +122,6 @@ pub struct LiveRange {
     pub uses_spill_weight_and_flags: u32,
 
     pub uses: UseList,
-
-    pub merged_into: LiveRangeIndex,
 }
 
 #[derive(Clone, Copy, Debug, PartialEq, Eq)]
@@ -197,8 +194,6 @@ impl Use {
     }
 }
 
-pub const SLOT_NONE: u8 = u8::MAX;
-
 #[derive(Clone, Debug)]
 pub struct LiveBundle {
     pub ranges: LiveRangeList,
@@ -208,7 +203,7 @@ pub struct LiveBundle {
     pub spill_weight_and_props: u32,
 }
 
-pub const BUNDLE_MAX_SPILL_WEIGHT: u32 = (1 << 28) - 1;
+pub const BUNDLE_MAX_SPILL_WEIGHT: u32 = (1 << 29) - 1;
 pub const MINIMAL_FIXED_BUNDLE_SPILL_WEIGHT: u32 = BUNDLE_MAX_SPILL_WEIGHT;
 pub const MINIMAL_BUNDLE_SPILL_WEIGHT: u32 = BUNDLE_MAX_SPILL_WEIGHT - 1;
 pub const BUNDLE_MAX_NORMAL_SPILL_WEIGHT: u32 = BUNDLE_MAX_SPILL_WEIGHT - 2;
@@ -221,14 +216,12 @@ impl LiveBundle {
         minimal: bool,
         fixed: bool,
         fixed_def: bool,
-        stack: bool,
     ) {
         debug_assert!(spill_weight <= BUNDLE_MAX_SPILL_WEIGHT);
         self.spill_weight_and_props = spill_weight
             | (if minimal { 1 << 31 } else { 0 })
             | (if fixed { 1 << 30 } else { 0 })
-            | (if fixed_def { 1 << 29 } else { 0 })
-            | (if stack { 1 << 28 } else { 0 });
+            | (if fixed_def { 1 << 29 } else { 0 });
     }
 
     #[inline(always)]
@@ -246,11 +239,6 @@ impl LiveBundle {
         self.spill_weight_and_props & (1 << 29) != 0
     }
 
-    #[inline(always)]
-    pub fn cached_stack(&self) -> bool {
-        self.spill_weight_and_props & (1 << 28) != 0
-    }
-
     #[inline(always)]
     pub fn set_cached_fixed(&mut self) {
         self.spill_weight_and_props |= 1 << 30;
@@ -261,14 +249,9 @@ impl LiveBundle {
         self.spill_weight_and_props |= 1 << 29;
     }
 
-    #[inline(always)]
-    pub fn set_cached_stack(&mut self) {
-        self.spill_weight_and_props |= 1 << 28;
-    }
-
     #[inline(always)]
     pub fn cached_spill_weight(&self) -> u32 {
-        self.spill_weight_and_props & ((1 << 28) - 1)
+        self.spill_weight_and_props & BUNDLE_MAX_SPILL_WEIGHT
     }
 }
 
@@ -302,7 +285,6 @@ pub struct SpillSet {
     pub class: RegClass,
     pub spill_bundle: LiveBundleIndex,
     pub required: bool,
-    pub size: u8,
     pub splits: u8,
 
     /// The aggregate [`CodeRange`] of all involved [`LiveRange`]s. The effect of this abstraction
@@ -318,7 +300,6 @@ pub(crate) const MAX_SPLITS_PER_SPILLSET: u8 = 2;
 pub struct VRegData {
     pub ranges: LiveRangeList,
     pub blockparam: Block,
-    pub is_ref: bool,
     // We don't initially know the RegClass until we observe a use of the VReg.
     pub class: Option<RegClass>,
 }
@@ -398,6 +379,55 @@ impl BlockparamIn {
     }
 }
 
+impl LiveRanges {
+    pub fn add(&mut self, range: CodeRange) -> LiveRangeIndex {
+        self.push(LiveRange {
+            range,
+            vreg: VRegIndex::invalid(),
+            bundle: LiveBundleIndex::invalid(),
+            uses_spill_weight_and_flags: 0,
+
+            uses: smallvec![],
+        })
+    }
+}
+
+impl LiveBundles {
+    pub fn add(&mut self) -> LiveBundleIndex {
+        self.push(LiveBundle {
+            allocation: Allocation::none(),
+            ranges: smallvec![],
+            spillset: SpillSetIndex::invalid(),
+            prio: 0,
+            spill_weight_and_props: 0,
+        })
+    }
+}
+
+impl VRegs {
+    pub fn add(&mut self, reg: VReg, data: VRegData) -> VRegIndex {
+        let idx = self.push(data);
+        debug_assert_eq!(reg.vreg(), idx.index());
+        idx
+    }
+}
+
+impl core::ops::Index<VReg> for VRegs {
+    type Output = VRegData;
+
+    #[inline(always)]
+    fn index(&self, idx: VReg) -> &Self::Output {
+        &self.storage[idx.vreg()]
+    }
+}
+
+impl core::ops::IndexMut<VReg> for VRegs {
+    #[inline(always)]
+    fn index_mut(&mut self, idx: VReg) -> &mut Self::Output {
+        &mut self.storage[idx.vreg()]
+    }
+}
+
 #[derive(Clone, Debug)]
 pub struct Env<'a, F: Function> {
     pub func: &'a F,
@@ -408,18 +438,16 @@ pub struct Env<'a, F: Function> {
     pub blockparam_outs: Vec<BlockparamOut>,
     pub blockparam_ins: Vec<BlockparamIn>,
 
-    pub ranges: Vec<LiveRange>,
-    pub bundles: Vec<LiveBundle>,
-    pub spillsets: Vec<SpillSet>,
-    pub vregs: Vec<VRegData>,
+    pub ranges: LiveRanges,
+    pub bundles: LiveBundles,
+    pub spillsets: SpillSets,
+    pub vregs: VRegs,
     pub pregs: Vec<PRegData>,
     pub allocation_queue: PrioQueue,
-    pub safepoints: Vec<Inst>, // Sorted list of safepoint insts.
-    pub safepoints_per_vreg: HashMap<usize, HashSet<Inst>>,
 
     pub spilled_bundles: Vec<LiveBundleIndex>,
     pub spillslots: Vec<SpillSlotData>,
-    pub slots_by_size: Vec<SpillSlotList>,
+    pub slots_by_class: [SpillSlotList; 3],
 
     pub extra_spillslots_by_class: [SmallVec<[Allocation; 2]>; 3],
     pub preferred_victim_by_class: [PReg; 3],
@@ -435,11 +463,9 @@ pub struct Env<'a, F: Function> {
     pub multi_fixed_reg_fixups: Vec<MultiFixedRegFixup>,
 
     // Output:
-    pub edits: Vec<(PosWithPrio, Edit)>,
     pub allocs: Vec<Allocation>,
     pub inst_alloc_offsets: Vec<u32>,
     pub num_spillslots: u32,
-    pub safepoint_slots: Vec<(ProgPoint, Allocation)>,
     pub debug_locations: Vec<(u32, ProgPoint, ProgPoint, Allocation)>,
 
     pub allocated_bundle_count: usize,
@@ -460,7 +486,7 @@ impl<'a, F: Function> Env<'a, F> {
     /// Get the VReg (with bundled RegClass) from a vreg index.
     #[inline]
     pub fn vreg(&self, index: VRegIndex) -> VReg {
-        let class = self.vregs[index.index()]
+        let class = self.vregs[index]
             .class
             .expect("trying to get a VReg before observing its class");
         VReg::new(index.index(), class)
@@ -469,7 +495,7 @@ impl<'a, F: Function> Env<'a, F> {
     /// Record the class of a VReg. We learn this only when we observe
     /// the VRegs in use.
     pub fn observe_vreg_class(&mut self, vreg: VReg) {
-        let old_class = self.vregs[vreg.vreg()].class.replace(vreg.class());
+        let old_class = self.vregs[vreg].class.replace(vreg.class());
         // We should never observe two different classes for two
         // mentions of a VReg in the source program.
         debug_assert!(old_class == None || old_class == Some(vreg.class()));
@@ -477,7 +503,7 @@ impl<'a, F: Function> Env<'a, F> {
 
     /// Is this vreg actually used in the source program?
     pub fn is_vreg_used(&self, index: VRegIndex) -> bool {
-        self.vregs[index.index()].class.is_some()
+        self.vregs[index].class.is_some()
     }
 }
 
@@ -508,6 +534,13 @@ pub struct SpillSlotList {
 }
 
 impl SpillSlotList {
+    pub fn new() -> Self {
+        SpillSlotList {
+            slots: smallvec![],
+            probe_start: 0,
+        }
+    }
+
     /// Get the next spillslot index in probing order, wrapping around
     /// at the end of the slots list.
     pub(crate) fn next_index(&self, index: usize) -> usize {
@@ -694,6 +727,51 @@ impl InsertedMoves {
     }
 }
 
+#[derive(Clone, Debug)]
+pub struct Edits {
+    edits: Vec<(PosWithPrio, Edit)>,
+}
+
+impl Edits {
+    #[inline(always)]
+    pub fn with_capacity(n: usize) -> Self {
+        Self {
+            edits: Vec::with_capacity(n),
+        }
+    }
+
+    #[inline(always)]
+    pub fn len(&self) -> usize {
+        self.edits.len()
+    }
+
+    #[inline(always)]
+    pub fn iter(&self) -> impl Iterator<Item = &(PosWithPrio, Edit)> {
+        self.edits.iter()
+    }
+
+    #[inline(always)]
+    pub fn into_edits(self) -> impl Iterator<Item = (ProgPoint, Edit)> {
+        self.edits.into_iter().map(|(pos, edit)| (pos.pos, edit))
+    }
+
+    /// Sort edits by the combination of their program position and priority. This is a stable sort
+    /// to preserve the order of the moves the parallel move resolver inserts.
+    #[inline(always)]
+    pub fn sort(&mut self) {
+        self.edits.sort_by_key(|&(pos_prio, _)| pos_prio.key());
+    }
+
+    pub fn add(&mut self, pos_prio: PosWithPrio, from: Allocation, to: Allocation) {
+        if from != to {
+            if from.is_reg() && to.is_reg() {
+                debug_assert_eq!(from.as_reg().unwrap().class(), to.as_reg().unwrap().class());
+            }
+            self.edits.push((pos_prio, Edit::Move { from, to }));
+        }
+    }
+}
+
 /// The fields in this struct are reversed in sort order so that the entire
 /// struct can be treated as a u64 for sorting purposes.
 #[derive(Clone, Copy, Debug, PartialEq, Eq)]
