@@ -14,11 +14,11 @@
 
 use super::{
     Env, InsertMovePrio, InsertedMove, InsertedMoves, LiveRangeFlag, LiveRangeIndex,
-    RedundantMoveEliminator, VRegIndex, SLOT_NONE,
+    RedundantMoveEliminator, VRegIndex,
 };
 use crate::ion::data_structures::{
-    u64_key, BlockparamIn, BlockparamOut, CodeRange, FixedRegFixupLevel, LiveRangeKey,
-    LiveRangeListEntry, PosWithPrio,
+    u64_key, BlockparamIn, BlockparamOut, CodeRange, Edits, FixedRegFixupLevel, LiveRangeKey,
+    LiveRangeListEntry,
 };
 use crate::ion::reg_traversal::RegTraversalIter;
 use crate::moves::{MoveAndScratchResolver, ParallelMoves};
@@ -53,9 +53,9 @@ impl<'a, F: Function> Env<'a, F> {
 
     pub fn get_alloc_for_range(&self, range: LiveRangeIndex) -> Allocation {
         trace!("get_alloc_for_range: {:?}", range);
-        let bundle = self.ranges[range.index()].bundle;
+        let bundle = self.ranges[range].bundle;
         trace!(" -> bundle: {:?}", bundle);
-        let bundledata = &self.bundles[bundle.index()];
+        let bundledata = &self.bundles[bundle];
         trace!(" -> allocation {:?}", bundledata.allocation);
         if bundledata.allocation != Allocation::none() {
             bundledata.allocation
@@ -63,9 +63,9 @@ impl<'a, F: Function> Env<'a, F> {
             trace!(" -> spillset {:?}", bundledata.spillset);
             trace!(
                 " -> spill slot {:?}",
-                self.spillsets[bundledata.spillset.index()].slot
+                self.spillsets[bundledata.spillset].slot
             );
-            self.spillslots[self.spillsets[bundledata.spillset.index()].slot.index()].alloc
+            self.spillslots[self.spillsets[bundledata.spillset].slot.index()].alloc
         }
     }
 
@@ -80,7 +80,7 @@ impl<'a, F: Function> Env<'a, F> {
         // sort VReg range lists and update with the final ranges.
         for vreg in &mut self.vregs {
             for entry in &mut vreg.ranges {
-                entry.range = self.ranges[entry.index.index()].range;
+                entry.range = self.ranges[entry.index].range;
             }
             vreg.ranges.sort_unstable_by_key(|entry| entry.range.from);
         }
@@ -286,8 +286,8 @@ impl<'a, F: Function> Env<'a, F> {
             // `blockparam_outs`, which are sorted by (block, vreg),
             // to fill in allocations.
             let mut prev = PrevBuffer::new(blockparam_in_idx);
-            for range_idx in 0..self.vregs[vreg.index()].ranges.len() {
-                let entry = self.vregs[vreg.index()].ranges[range_idx];
+            for range_idx in 0..self.vregs[vreg].ranges.len() {
+                let entry = self.vregs[vreg].ranges[range_idx];
                 let alloc = self.get_alloc_for_range(entry.index);
                 let range = entry.range;
                 trace!(
@@ -307,7 +307,7 @@ impl<'a, F: Function> Env<'a, F> {
                             vreg.index(),
                             alloc,
                             entry.index.index(),
-                            self.ranges[entry.index.index()].bundle.raw_u32(),
+                            self.ranges[entry.index].bundle.raw_u32(),
                         ),
                     );
                     self.annotate(
@@ -317,7 +317,7 @@ impl<'a, F: Function> Env<'a, F> {
                             vreg.index(),
                             alloc,
                             entry.index.index(),
-                            self.ranges[entry.index.index()].bundle.raw_u32(),
+                            self.ranges[entry.index].bundle.raw_u32(),
                         ),
                     );
                 }
@@ -350,7 +350,7 @@ impl<'a, F: Function> Env<'a, F> {
 
                     if prev.range.to >= range.from
                         && (prev.range.to > range.from || !self.is_start_of_block(range.from))
-                        && !self.ranges[entry.index.index()].has_flag(LiveRangeFlag::StartsAtDef)
+                        && !self.ranges[entry.index].has_flag(LiveRangeFlag::StartsAtDef)
                     {
                         trace!(
                             "prev LR {} abuts LR {} in same block; moving {} -> {} for v{}",
@@ -556,18 +556,14 @@ impl<'a, F: Function> Env<'a, F> {
                 }
 
                 // Scan over def/uses and apply allocations.
-                for use_idx in 0..self.ranges[entry.index.index()].uses.len() {
-                    let usedata = self.ranges[entry.index.index()].uses[use_idx];
+                for use_idx in 0..self.ranges[entry.index].uses.len() {
+                    let usedata = self.ranges[entry.index].uses[use_idx];
                     trace!("applying to use: {:?}", usedata);
                     debug_assert!(range.contains_point(usedata.pos));
                     let inst = usedata.pos.inst();
                     let slot = usedata.slot;
                     let operand = usedata.operand;
-                    // Safepoints add virtual uses with no slots;
-                    // avoid these.
-                    if slot != SLOT_NONE {
-                        self.set_alloc(inst, slot as usize, alloc);
-                    }
+                    self.set_alloc(inst, slot as usize, alloc);
                     if let OperandConstraint::Reuse(_) = operand.constraint() {
                         reuse_input_insts.push(inst);
                     }
@@ -772,7 +768,7 @@ impl<'a, F: Function> Env<'a, F> {
         inserted_moves
     }
 
-    pub fn resolve_inserted_moves(&mut self, mut inserted_moves: InsertedMoves) {
+    pub fn resolve_inserted_moves(&mut self, mut inserted_moves: InsertedMoves) -> Edits {
         // For each program point, gather all moves together. Then
         // resolve (see cases below).
         let mut i = 0;
@@ -789,20 +785,13 @@ impl<'a, F: Function> Env<'a, F> {
             from: ProgPoint,
             to: ProgPoint,
         ) {
-            // If any safepoints in range, clear and return.
-            // Also, if we cross a block boundary, clear and return.
+            // If we cross a block boundary, clear and return.
             if this.cfginfo.insn_block[from.inst().index()]
                 != this.cfginfo.insn_block[to.inst().index()]
             {
                 redundant_moves.clear();
                 return;
             }
-            for inst in from.inst().index()..=to.inst().index() {
-                if this.func.requires_refs_on_stack(Inst::new(inst)) {
-                    redundant_moves.clear();
-                    return;
-                }
-            }
 
             let start_inst = if from.pos() == InstPosition::Before {
                 from.inst()
@@ -839,6 +828,7 @@ impl<'a, F: Function> Env<'a, F> {
         }
 
         let mut last_pos = ProgPoint::before(Inst::new(0));
+        let mut edits = Edits::with_capacity(self.func.num_insts());
 
         while i < inserted_moves.moves.len() {
             let start = i;
@@ -903,12 +893,15 @@ impl<'a, F: Function> Env<'a, F> {
                     0,
                     None,
                 );
+                let mut dedicated_scratch = self.env.scratch_by_class[regclass as usize];
                 let key = LiveRangeKey::from_range(&CodeRange {
                     from: pos_prio.pos,
                     to: pos_prio.pos.next(),
                 });
-                let get_reg = || {
-                    if let Some(reg) = self.env.scratch_by_class[regclass as usize] {
+                let find_free_reg = || {
+                    // Use the dedicated scratch register first if it is
+                    // available.
+                    if let Some(reg) = dedicated_scratch.take() {
                         return Some(Allocation::reg(reg));
                     }
                     while let Some(preg) = scratch_iter.next() {
@@ -954,12 +947,12 @@ impl<'a, F: Function> Env<'a, F> {
                 };
                 let preferred_victim = self.preferred_victim_by_class[regclass as usize];
 
-                let scratch_resolver = MoveAndScratchResolver::new(
-                    get_reg,
+                let scratch_resolver = MoveAndScratchResolver {
+                    find_free_reg,
                     get_stackslot,
                     is_stack_alloc,
-                    preferred_victim,
-                );
+                    borrowed_scratch_reg: preferred_victim,
+                };
 
                 let resolved = scratch_resolver.compute(resolved);
 
@@ -982,7 +975,7 @@ impl<'a, F: Function> Env<'a, F> {
                     trace!("  resolved: {} -> {} ({:?})", src, dst, to_vreg);
                     let action = redundant_moves.process_move(src, dst, to_vreg);
                     if !action.elide {
-                        self.add_move_edit(pos_prio, src, dst);
+                        edits.add(pos_prio, src, dst);
                     } else {
                         trace!("    -> redundant move elided");
                     }
@@ -994,13 +987,12 @@ impl<'a, F: Function> Env<'a, F> {
         // be a stable sort! We have to keep the order produced by the
         // parallel-move resolver for all moves within a single sort
         // key.
-        self.edits.sort_by_key(|&(pos_prio, _)| pos_prio.key());
-        self.stats.edits_count = self.edits.len();
+        edits.sort();
+        self.stats.edits_count = edits.len();
 
         // Add debug annotations.
         if self.annotations_enabled {
-            for i in 0..self.edits.len() {
-                let &(pos_prio, ref edit) = &self.edits[i];
+            for &(pos_prio, ref edit) in edits.iter() {
                 match edit {
                     &Edit::Move { from, to } => {
                         self.annotate(pos_prio.pos, format!("move {} -> {}", from, to));
@@ -1008,14 +1000,7 @@ impl<'a, F: Function> Env<'a, F> {
                 }
             }
         }
-    }
 
-    pub fn add_move_edit(&mut self, pos_prio: PosWithPrio, from: Allocation, to: Allocation) {
-        if from != to {
-            if from.is_reg() && to.is_reg() {
-                debug_assert_eq!(from.as_reg().unwrap().class(), to.as_reg().unwrap().class());
-            }
-            self.edits.push((pos_prio, Edit::Move { from, to }));
-        }
+        edits
     }
 }
