@@ -265,7 +265,6 @@ impl<'a, F: Function> Env<'a, F> {
         let minimal;
         let mut fixed = false;
         let mut fixed_def = false;
-        let mut stack = false;
         let bundledata = &self.bundles[bundle];
         let first_range = bundledata.ranges[0].index;
         let first_range_data = &self.ranges[first_range];
@@ -286,12 +285,7 @@ impl<'a, F: Function> Env<'a, F> {
                         trace!("  -> is fixed def");
                         fixed_def = true;
                     }
-                }
-                if let OperandConstraint::Stack = u.operand.constraint() {
-                    trace!("  -> stack operand at {:?}: {:?}", u.pos, u.operand);
-                    stack = true;
-                }
-                if stack && fixed {
+
                     break;
                 }
             }
@@ -343,7 +337,6 @@ impl<'a, F: Function> Env<'a, F> {
             minimal,
             fixed,
             fixed_def,
-            stack,
         );
     }
 
@@ -1056,12 +1049,6 @@ impl<'a, F: Function> Env<'a, F> {
             let fixed_preg = match req {
                 Requirement::FixedReg(preg) | Requirement::FixedStack(preg) => Some(preg),
                 Requirement::Register => None,
-                Requirement::Stack => {
-                    // If we must be on the stack, mark our spillset
-                    // as required immediately.
-                    self.spillsets[self.bundles[bundle].spillset].required = true;
-                    return Ok(());
-                }
 
                 Requirement::Any => {
                     self.spilled_bundles.push(bundle);
